% Flyback converter reverse engineering output ripple calculations
% By K.B. October 24th, 2023
% Info:
% 1)
% This program implements function for each phase of converter's working cycle:
% - "ON" (MOSFET swithced “ON”, primary current increases to its threshold level, energy stored in magnetic field)
%    str_vals = calc_ON(str_vals)
% - "OFF1" (Secondary coil gets energised and then demagnetizes, output capacitors receive charge.
%    Reflected voltage across auxilary coil switches "OFF" MOSFET)
%    str_vals = calc_OFF1(str_vals)
% - "FEEDBACK" (Shunt regulator switched "ON", Output capacitors discharges.
%    C2 discharge path to ground activated by feedback signal, so MOSFET's gate voltage not rising.)
%    str_vals = calc_FB(str_vals)
% - "OFF2" (Shunt regulator switched "OFF", C2 capacotor recharges and MOSFET's gate voltage rises back to threshold level)
%    str_vals = calc_OFF2(str_vals)
% 2)
% Functions calculates only initial and final values of parameters for corresponding phase
% 3)
% All the initial parameters and calculated values stored in structure str_vals.
% Functions receive this structure as argument, then updates stored values
% after calculations.
% 4)
% Functions perform one after one in loop, and "FEEDBACK" function performs
% only when calculated output voltage exceeds threshold level.

clear all
clc

% Component/execution parameters
% Pack into tructure
str_vals.C2 = 0.022*10^-6;       % F
str_vals.EC3 = 330*10^-6;        % F
str_vals.EC4 = 470*10^-6;        % F
str_vals.Lp = 997*10^-6;         % H
str_vals.Naux = 6;               % Turns
str_vals.Np = 72;                % Turns
str_vals.Ns1 = 15;               % Turns
str_vals.Ns2 = 7;                % Turns
str_vals.R1 = 1000000;           % Ohm
str_vals.R2 = 1000;              % Ohm
str_vals.R5 = 220;               % Ohm
str_vals.R6 = 0.47;              % Ohm
str_vals.R7 = 1000000;           % Ohm
str_vals.R11 = 3900;             % Ohm
str_vals.R13 = 3000;             % Ohm
str_vals.R14 = 49900;            % Ohm
str_vals.RDS = 2.5;              % Ohm, MOSFET "ON" DS resistance
str_vals.RLp = 0.7;              % Ohm, Primary coil resistance.
str_vals.RLs = 0.6;              % Ohm, eecondary coil resistance.
str_vals.Vbe_sat = 0.7;          % V
str_vals.VDC = 310;              % V
str_vals.VDW = 18.6;             % V, Zener diodes clamp voltage
str_vals.VGS_Th = 4;             % V
str_vals.VK = 2;                 % V, Cathode of shunt regulator
str_vals.VLed_F = 1.3;           % V, Led forvard voltage
str_vals.Vref_Min = 2.298;       % V, [INPUT}
str_vals.Vref_Typ = 2.495;       % V
str_vals.Vref_Max = 2.545;       % V, [INPUT}

str_vals.eff = 1;                % Efficiency of energy transfer between primary and secondary coil
str_vals.ptsNum = 100;           % number of time points for numeric calculations
str_vals.cycleNum = 21;          % number of flyback working cycles to simulate [INPUT}


% ---------------------------------------------
% Operational parameters

% EC3 and EC4 output equivelent capacitance
str_vals.EC = str_vals.EC3*str_vals.EC4/(str_vals.EC3+str_vals.EC4);

% Output threshold voltage (Voltage across output EC capacitors controlled by Shunt regulator)
str_vals.V_EC_Th_Min = str_vals.Vref_Min*(str_vals.R14+str_vals.R11)/str_vals.R11;
str_vals.V_EC_Th = str_vals.Vref_Typ*(str_vals.R14+str_vals.R11)/str_vals.R11;
str_vals.V_EC_Th_Max = str_vals.Vref_Max*(str_vals.R14+str_vals.R11)/str_vals.R11;

% Secondary coil effective inductance
str_vals.Ls = str_vals.Lp/(str_vals.Np/(str_vals.Ns1+str_vals.Ns2))^2;

% Total resistance of primaty coil loop when MOSFET is "ON"
str_vals.RP = str_vals.RLp + str_vals.RDS + str_vals.R6;

% Maximum current in primaty coil loop
str_vals.I_Lp_Max = str_vals.VDC/str_vals.RP;

% Threshold current in primaty coil loop after which Q3 swithces ON
str_vals.I_Lp_Th = str_vals.Vbe_sat/str_vals.R6;

% Time constant for primary coil
str_vals.tau_Lp = str_vals.Lp/str_vals.RP;

% C2 time constants
% A - Zeners - "OFF", Q2,Q3 - "OFF"
% B - Zeners - "ON", Q2,Q3 - "OFF"
% C - Zeners - "OFF", Q2,Q3 - "ON"
str_vals.tau_C2A = (str_vals.R1+str_vals.R7+str_vals.R2)*str_vals.C2;
str_vals.tau_C2B = str_vals.R2*str_vals.C2;
str_vals.tau_C2C = (str_vals.R2+str_vals.R5+str_vals.R6)*str_vals.C2;

% Output ECs capacitors time constants
% "VD" - discharge via R11,R14 voltage divider
% "Led" - discharge via R13 and LED
str_vals.tau_VD = (str_vals.R11+str_vals.R14)*str_vals.EC;
str_vals.tau_Led = str_vals.R13*str_vals.EC;

% ---------------------------------------------
% Initial values to start imulation
% Pack into structure
str_vals.deltaE = 0;
str_vals.deltaQ = 0;
str_vals.deltaV = 0;
str_vals.TFB = 0;
str_vals.TON = 0;
str_vals.TOFF1 = 0;
str_vals.TOFF2 = 0;
str_vals.I_Lp_i = 0;
str_vals.I_Lp_f = 0;
str_vals.I_Ls_i = 0;
str_vals.I_Ls_f = 0;
str_vals.V_C2_i = 0;
str_vals.V_C2_f = -str_vals.VGS_Th;    % C2 Just charged to MOSFET's "ON" threshold voltage
str_vals.V_EC_i = 0;
str_vals.V_EC_f = str_vals.V_EC_Th_Min;   % Assume ECs already charged to this level
str_vals.V_Laux_i = 0;
str_vals.V_Laux_f = 0;
str_vals.V_Lp_i = 0;
str_vals.V_Lp_f = 0;
str_vals.V_Ls_i = 0;
str_vals.V_Ls_f = 0;
str_vals.VGS_i = 0;
str_vals.VGS_f = -str_vals.V_C2_f;
str_vals.VTR2_i = 0;
str_vals.VTR2_f = -str_vals.V_Laux_f;

str_vals.E_EC_i = 0;
str_vals.E_EC_f = 0.5*str_vals.EC*(str_vals.V_EC_Th_Min)^2;
str_vals.Q_EC_i = 0;
str_vals.Q_EC_f = str_vals.EC*str_vals.V_EC_Th_Min;


% ---------------------------------------------
% Functions
function str_vals = calc_ON(str_vals)

  % ---------------------------------------------
  % "ON" Period (Primary current increases to its threshold level)
  % Duration: TON;
  % initial time moment: ton_i;
  % final moment: ton_f;

  % Unpack component/execution parameters
  EC = str_vals.EC;
  I_Lp_Max = str_vals.I_Lp_Max;
  I_Lp_Th = str_vals.I_Lp_Th;
  Naux = str_vals.Naux;
  Np = str_vals.Np;
  tau_C2B = str_vals.tau_C2B;
  tau_Lp = str_vals.tau_Lp;
  tau_VD = str_vals.tau_VD;
  VDC = str_vals.VDC;
  VDW = str_vals.VDW;
  ptsNum = str_vals.ptsNum;


  % Receive values from the end of previous phase that not changed
  % at the beginning of this phase
  E_EC_ton_i = str_vals.E_EC_f;
  I_Lp_ton_i = str_vals.I_Lp_f;
  I_Ls_ton_i = str_vals.I_Ls_f;
  Q_EC_ton_i = str_vals.Q_EC_f;
  V_C2_ton_i = str_vals.V_C2_f;
  V_EC_ton_i = str_vals.V_EC_f;

  % Known\definite values
  I_Lp_ton_f = I_Lp_Th;     % By definition
  I_Ls_ton_f = I_Ls_ton_i;  % By definition
  V_Ls_ton_i = 0;           % Ignore
  V_Ls_ton_f = 0;           % Ignore

  % Calculations
  V_Lp_ton_i = -VDC;
  V_Laux_ton_i = (Naux/Np)*V_Lp_ton_i;
  VTR2_ton_i = -V_Laux_ton_i;
  TON = tau_Lp*log((I_Lp_ton_i-I_Lp_Max)/(I_Lp_ton_f-I_Lp_Max));
  V_Lp_ton_f = V_Lp_ton_i*exp(-TON/tau_Lp);
  V_Laux_ton_f = (Naux/Np)*V_Lp_ton_f;
  VTR2_ton_f = -V_Laux_ton_f;

  % Calculation of C2 final voltage
  tspan = linspace(0,TON,ptsNum);
  % Differential equation for C2 voltage
  % (in a form of anonymus function in a form necessary for ode45 solver)
  dvdt = @(t,v) (1/tau_C2B)*(-v+VTR2_ton_i*exp(-t/tau_Lp)-VDW);
  % Solved using ode45 solver for time period equivalent TON in lenght
  [T,V] = ode45(dvdt, tspan, V_C2_ton_i);
  % Voltage of C2 at the end of TON period
  % is last element of array of voltages calculated for time span
  % length of TON
  V_C2_ton_f = V(ptsNum,1);

  if (VTR2_ton_i>VDW)
    VGS_ton_i = VDW;
  else
    VGS_ton_i = -V_C2_ton_i;
  endif

  if (VTR2_ton_f>VDW)
    VGS_ton_f = VDW;
  else
    VGS_ton_f = -V_C2_ton_f;
  endif

  % Calc stored energy deltaE
  p_Lp = @(t) VDC.*exp(-t./tau_Lp).*I_Lp_Max.*(1-exp(-t./tau_Lp));
  % Integration
  deltaE = integral(p_Lp, 0, TON);

  % Output EC capacitors V, E, Q
  V_EC_ton_f = V_EC_ton_i*exp(-TON/tau_VD);
  E_EC_ton_f = 0.5*EC*V_EC_ton_f^2;
  Q_EC_ton_f = EC*V_EC_ton_f;


  % Update calculated values back to structure
  str_vals.deltaE = deltaE;
  str_vals.deltaQ = 0;
  str_vals.deltaV = 0;
  str_vals.TFB = 0;
  str_vals.TON = TON;
  str_vals.TOFF1 = 0;
  str_vals.TOFF2 = 0;
  str_vals.E_EC_i = E_EC_ton_i;
  str_vals.E_EC_f = E_EC_ton_f;
  str_vals.I_Lp_i = I_Lp_ton_i;
  str_vals.I_Lp_f = I_Lp_ton_f;
  str_vals.I_Ls_i = I_Ls_ton_i;
  str_vals.I_Ls_f = I_Ls_ton_f;
  str_vals.Q_EC_i = Q_EC_ton_i;
  str_vals.Q_EC_f = Q_EC_ton_f;
  str_vals.V_C2_i = V_C2_ton_i;
  str_vals.V_C2_f = V_C2_ton_f;
  str_vals.V_EC_i = V_EC_ton_i;
  str_vals.V_EC_f = V_EC_ton_f;
  str_vals.V_Laux_i = V_Laux_ton_i;
  str_vals.V_Laux_f = V_Laux_ton_f;
  str_vals.V_Lp_i = V_Lp_ton_i;
  str_vals.V_Lp_f = V_Lp_ton_f;
  str_vals.V_Ls_i = V_Ls_ton_i;
  str_vals.V_Ls_f = V_Ls_ton_f;
  str_vals.VGS_i = VGS_ton_i;
  str_vals.VGS_f = VGS_ton_f;
  str_vals.VTR2_i = VTR2_ton_i;
  str_vals.VTR2_f = VTR2_ton_f;
endfunction

function str_vals = calc_OFF1(str_vals)

  % ---------------------------------------------
  % "OFF1" Period (Secondary coil demagnetize, EC gets charge)
  % Duration: TOFF1;
  % initial time moment: toff1_i;
  % final moment: toff1_f;

  % Unpack component/execution parameters
  EC = str_vals.EC;
  Ls = str_vals.Ls;
  Naux = str_vals.Naux;
  Ns1 = str_vals.Ns1;
  Ns2 = str_vals.Ns2;
  RLs = str_vals.RLs;
  VDC = str_vals.VDC;
  VDW = str_vals.VDW;
  ptsNum = str_vals.ptsNum;

  % Unpack operational parameters
  tau_C2A = str_vals.tau_C2A;

  % Receive values from the end of previous phase that not changed
  % at the beginning of this phase
  deltaE = str_vals.deltaE*str_vals.eff;
  E_EC_toff1_i = str_vals.E_EC_f;
  Q_EC_toff1_i = str_vals.Q_EC_f;
  V_C2_toff1_i = str_vals.V_C2_f;
  V_EC_toff1_i = str_vals.V_EC_f;


  % Known\definite values
  I_Lp_toff1_i = 0;   % Ignoring
  I_Lp_toff1_f = 0;   % Ignoring
  I_Ls_toff1_f = 0;   % By definition
  V_Lp_toff1_i = 0;   % Ignoring
  V_Lp_toff1_f = 0;   % Ignoring
  V_Ls_toff1_f = 0;   % By definition
  V_Laux_toff1_f = (Naux/(Ns1+Ns2))*V_Ls_toff1_f;
  VTR2_toff1_f = -V_Laux_toff1_f;

  % Output EC capacitors V, E, Q
  E_EC_toff1_f = E_EC_toff1_i + deltaE;
  V_EC_toff1_f = sqrt(2*E_EC_toff1_f/EC);
  Q_EC_toff1_f = V_EC_toff1_f*EC;
  deltaV = V_EC_toff1_f - V_EC_toff1_i;
  deltaQ = Q_EC_toff1_f - Q_EC_toff1_i;

  I_Ls_toff1_i = sqrt(2*deltaE/Ls);

  % Calc TOFF1
  DIDT = -V_EC_toff1_i/Ls;
  w0 = 1/sqrt(Ls*EC);
  a = RLs/(2*Ls);
  b = sqrt(a^2-w0^2);
  A1 = (-(V_EC_toff1_i/Ls)+(a+b)*I_Ls_toff1_i)/(2*b);
  A2 = I_Ls_toff1_i-A1;
  % Function to integrate
  i_t = @(t) A1*exp((-a+b)*t) + A2*exp((-a-b)*t);
  % Integral equation "deltaQ = integral of i_t from 0 to T"
  % in a form suitable for solver fsolve
  Q = @(T) (integral(i_t, 0, T)-deltaQ);
  % Initial gues need for solver fsolve
  initGuess = 1*10^-6;
  % Solution
  TOFF1 = fsolve (Q,initGuess);

  V_Ls_toff1_i = Ls*(I_Ls_toff1_i-I_Ls_toff1_f)/TOFF1;
  V_Laux_toff1_i = (Naux/(Ns1+Ns2))*V_Ls_toff1_i;
  VTR2_toff1_i = -V_Laux_toff1_i;

  % Calc C2 Voltage
  % time points (need for numeric calculations)
  tspan = linspace(0,TOFF1,ptsNum);
  % Differential equation for C2 voltage
  % (in a form of anonymus function in a form necessary for ode45 solver)
  dvdt = @(t,v) (1/tau_C2A)*(-v+VTR2_toff1_i*(1-t/TOFF1)-VDC);
  % Solved using ode45 solver for time period equivalent TOFF1 in lenght
  [T,V] = ode45(dvdt, tspan, V_C2_toff1_i);
  % Voltage of C2 at the end of TON period
  % is last element of array of voltages calculated for time span
  % length of TOFF1
  V_C2_toff1_f = V(ptsNum,1);

  %VGS_toff1_i = VTR2_toff1_i;
  %VGS_toff1_f = V_C2_toff1_f;
  if (VTR2_toff1_i>VDW)
    VGS_toff1_i = VDW;
  else
    VGS_toff1_i = -V_C2_toff1_i;
  endif

  if (VTR2_toff1_f>VDW)
    VGS_toff1_f = VDW;
  else
    VGS_toff1_f = -V_C2_toff1_f;
  endif

  % Update calculated values back to structure
  str_vals.deltaE = deltaE;
  str_vals.deltaQ = deltaQ;
  str_vals.deltaV = deltaV;
  str_vals.TFB = 0;
  str_vals.TON = 0;
  str_vals.TOFF1 = TOFF1;
  str_vals.TOFF2 = 0;
  str_vals.E_EC_i = E_EC_toff1_i;
  str_vals.E_EC_f = E_EC_toff1_f;
  str_vals.I_Lp_i = I_Lp_toff1_i;
  str_vals.I_Lp_f = I_Lp_toff1_f;
  str_vals.I_Ls_i = I_Ls_toff1_i;
  str_vals.I_Ls_f = I_Ls_toff1_f;
  str_vals.Q_EC_i = Q_EC_toff1_i;
  str_vals.Q_EC_f = Q_EC_toff1_f;
  str_vals.V_C2_i = V_C2_toff1_i;
  str_vals.V_C2_f = V_C2_toff1_f;
  str_vals.V_EC_i = V_EC_toff1_i;
  str_vals.V_EC_f = V_EC_toff1_f;
  str_vals.V_Laux_i = V_Laux_toff1_i;
  str_vals.V_Laux_f = V_Laux_toff1_f;
  str_vals.V_Lp_i = V_Lp_toff1_i;
  str_vals.V_Lp_f = V_Lp_toff1_f;
  str_vals.V_Ls_i = V_Ls_toff1_i;
  str_vals.V_Ls_f = V_Ls_toff1_f;
  str_vals.VGS_i = VGS_toff1_i;
  str_vals.VGS_f = VGS_toff1_f;
  str_vals.VTR2_i = VTR2_toff1_i;
  str_vals.VTR2_f = VTR2_toff1_f;

endfunction


function str_vals = calc_FB(str_vals)

  % ---------------------------------------------
  % "FEEDBACK" Period (EC output capacitors discharges back to threshold level)
  % Duration: TFB;
  % initial time moment: tfb_i;
  % final moment: tfb_f;

  % Unpack component/execution parameters
  EC = str_vals.EC;
  R11 = str_vals.R11;
  R13 = str_vals.R13;
  R14 = str_vals.R14;
  tau_C2C = str_vals.tau_C2C;
  tau_Led = str_vals.tau_Led;
  tau_VD = str_vals.tau_VD;
  VDW = str_vals.VDW;
  VLed_F = str_vals.VLed_F;
  VK = str_vals.VLed_F;
  V_EC_Th_Min = str_vals.V_EC_Th_Min;



  % Initial values unchanged at rhe beginning of this period
  E_EC_tfb_i = str_vals.E_EC_f;
  Q_EC_tfb_i = str_vals.Q_EC_f;
  V_EC_tfb_i = str_vals.V_EC_f;
  V_C2_tfb_i = str_vals.V_C2_f;

  % Initial values not changed during this period
  I_Lp_tfb_i = 0;   % Ignore
  I_Lp_tfb_f = 0;   % Ignore
  I_Ls_tfb_i = 0;   % Ignore
  I_Ls_tfb_f = 0;   % Ignore
  V_Laux_tfb_i = 0;   % Ignore
  V_Laux_tfb_f = 0;   % Ignore
  V_Lp_tfb_i = 0;   % Ignore
  V_Lp_tfb_f = 0;   % Ignore
  V_Ls_tfb_i = 0;   % Ignore
  V_Ls_tfb_f = 0;   % Ignore
  VTR2_tfb_i = -V_Ls_tfb_i;
  VTR2_tfb_f = -V_Ls_tfb_f;

  % Calculate extra charge acumulated in ECs above threhold value
  Q_EC_Th_Min = EC*V_EC_Th_Min;
  extraQ = Q_EC_tfb_i - Q_EC_Th_Min;


  I_VD_tfb_i = V_EC_tfb_i/(R11+R14);
  I_Led_tfb_i = (V_EC_tfb_i-VLed_F-VK)/R13;

  % Calc TFB
  % Function to integrate
  i_t = @(t) I_VD_tfb_i*exp(-t/tau_VD) + I_Led_tfb_i*exp(-t/tau_Led);
  % Integral equation "deltaQ = integral of i_t from 0 to T"
  % in a form suitable for solver fsolve
  Q = @(T) (integral(i_t, 0, T)-extraQ);
  % Initial gues need for solver fsolve
  initGuess = 1*10^-6;
  % Solution
  TFB = fsolve (Q,initGuess);

  % Output EC capacitors V, E, Q
  Q_EC_tfb_f = Q_EC_tfb_i-extraQ;
  V_EC_tfb_f = Q_EC_tfb_f/EC;
  E_EC_tfb_f = 0.5*EC*V_EC_tfb_f^2;


  V_C2_tfb_f = V_C2_tfb_i*exp(-TFB/tau_C2C);
  %VGS_tfb_i = -V_C2_tfb_i;
  %VGS_tfb_f = -V_C2_tfb_f;

  if (VTR2_tfb_i>VDW)
    VGS_tfb_i = VDW;
  else
    VGS_tfb_i = -V_C2_tfb_i;
  endif

  if (VTR2_tfb_f>VDW)
    VGS_tfb_f = VDW;
  else
    VGS_tfb_f = -V_C2_tfb_f;
  endif

  % Update calculated values back to structure
  str_vals.deltaE = 0;
  str_vals.deltaQ = 0;
  str_vals.deltaV = 0;
  str_vals.TFB = TFB;
  str_vals.TON = 0;
  str_vals.TOFF1 = 0;
  str_vals.TOFF2 = 0;
  str_vals.E_EC_i = E_EC_tfb_i;
  str_vals.E_EC_f = E_EC_tfb_f;
  str_vals.I_Lp_i = I_Lp_tfb_i;
  str_vals.I_Lp_f = I_Lp_tfb_f;
  str_vals.I_Ls_i = I_Ls_tfb_i;
  str_vals.I_Ls_f = I_Ls_tfb_f;
  str_vals.Q_EC_i = Q_EC_tfb_i;
  str_vals.Q_EC_f = Q_EC_tfb_f;
  str_vals.V_C2_i = V_C2_tfb_i;
  str_vals.V_C2_f = V_C2_tfb_f;
  str_vals.V_EC_i = V_EC_tfb_i;
  str_vals.V_EC_f = V_EC_tfb_f;
  str_vals.V_Laux_i = V_Laux_tfb_i;
  str_vals.V_Laux_f = V_Laux_tfb_f;
  str_vals.V_Lp_i = V_Lp_tfb_i;
  str_vals.V_Lp_f = V_Lp_tfb_f;
  str_vals.V_Ls_i = V_Ls_tfb_i;
  str_vals.V_Ls_f = V_Ls_tfb_f;
  str_vals.VGS_i = VGS_tfb_i;
  str_vals.VGS_f = VGS_tfb_f;
  str_vals.VTR2_i = VTR2_tfb_i;
  str_vals.VTR2_f = VTR2_tfb_f;

endfunction


function str_vals = calc_OFF2(str_vals)


  % ---------------------------------------------
  % "OFF2" Period (EC output capacitors discharges back to threshold level)
  % Duration: TOFF2;
  % initial time moment: toff2_i;
  % final moment: toff2_f;

  % Unpack values
  EC = str_vals.EC;
  tau_C2A = str_vals.tau_C2A;
  tau_VD = str_vals.tau_VD;
  VDC = str_vals.VDC;
  VGS_Th = str_vals.VGS_Th;

  % Initial values unchanged at rhe beginning of this period
  E_EC_toff2_i = str_vals.E_EC_f;
  Q_EC_toff2_i = str_vals.Q_EC_f;
  V_C2_toff2_i = str_vals.V_C2_f;
  V_EC_toff2_i = str_vals.V_EC_f;

  V_C2_toff2_f = -VGS_Th; % By definition

  % Initial values not changed during this period
  % Initial values not changed during this period
  I_Lp_toff2_i = 0;   % By definition
  I_Lp_toff2_f = 0;   % By definition
  I_Ls_toff2_i = 0;   % By definition
  I_Ls_toff2_f = 0;   % By definition
  V_Laux_toff2_i = 0;   % By definition
  V_Laux_toff2_f = 0;   % By definitione
  V_Lp_toff2_i = 0;   % By definition
  V_Lp_toff2_f = 0;   % By definition
  V_Ls_toff2_i = 0;   % By definition
  V_Ls_toff2_f = 0;   % By definition
  VTR2_toff2_i = -V_Ls_toff2_i;
  VTR2_toff2_f = -V_Ls_toff2_f;

  TOFF2 = tau_C2A*log((V_C2_toff2_i+VDC)/(V_C2_toff2_f+VDC));

  V_EC_toff2_f = V_EC_toff2_i*exp(-TOFF2/tau_VD);
  E_EC_toff2_f = 0.5*EC*V_EC_toff2_f^2;
  Q_EC_toff2_f = EC*V_EC_toff2_f;

  VGS_toff2_i = -V_C2_toff2_i;
  VGS_toff2_f = -V_C2_toff2_f;

  % Update calculated values back to structure
  str_vals.deltaE = 0;
  str_vals.deltaQ = 0;
  str_vals.deltaV = 0;
  str_vals.TFB = 0;
  str_vals.TON = 0;
  str_vals.TOFF1 = 0;
  str_vals.TOFF2 = TOFF2;
  str_vals.E_EC_i = E_EC_toff2_i;
  str_vals.E_EC_f = E_EC_toff2_f;
  str_vals.I_Lp_i = I_Lp_toff2_i;
  str_vals.I_Lp_f = I_Lp_toff2_f;
  str_vals.I_Ls_i = I_Ls_toff2_i;
  str_vals.I_Ls_f = I_Ls_toff2_f;
  str_vals.Q_EC_i = Q_EC_toff2_i;
  str_vals.Q_EC_f = Q_EC_toff2_f;
  str_vals.V_C2_i = V_C2_toff2_i;
  str_vals.V_C2_f = V_C2_toff2_f;
  str_vals.V_EC_i = V_EC_toff2_i;
  str_vals.V_EC_f = V_EC_toff2_f;
  str_vals.V_Laux_i = V_Laux_toff2_i;
  str_vals.V_Laux_f = V_Laux_toff2_f;
  str_vals.V_Lp_i = V_Lp_toff2_i;
  str_vals.V_Lp_f = V_Lp_toff2_f;
  str_vals.V_Ls_i = V_Ls_toff2_i;
  str_vals.V_Ls_f = V_Ls_toff2_f;
  str_vals.VGS_i = VGS_toff2_i;
  str_vals.VGS_f = VGS_toff2_f;
  str_vals.VTR2_i = VTR2_toff2_i;
  str_vals.VTR2_f = VTR2_toff2_f;
endfunction




% ---------------------------------------------
% Simulation

% Initialisation of cycle
t=0;  % start time
currRow = 1;   % initial index for Vout ripple plot
plotData(currRow,:) = [t,str_vals.V_EC_f]; % Add first entry for Vout ripple plot
fbStartCoord = [0,0];   % initial coordinates of (optional) label on plot
% Cycle
for n = 1:str_vals.cycleNum

  % "ON" Period
  tstart = t;               % Start of current cycle time ("ON" initial time)
  resultData(n,1)=n;        % Save current cycle No. to table with results
  resultData(n,2)=tstart;   % Save current cycle start time to table with results
  str_vals = calc_ON(str_vals);                 % Calculate values for "ON" period
  currRow = currRow+1;                          % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_i];    % Add row for Vout ripple plot
  t = t+str_vals.TON;       % "ON" period final time
  currRow = currRow+1;      % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_f];   % Add row for Vout ripple plot
  resultData(n,3)=str_vals.TON; % Save TON value to table with results

  % "OFF1" Period
  str_vals = calc_OFF1(str_vals); % Calculate values for "OFF1N" period
  currRow = currRow+1;            % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_i];  % Add row for Vout ripple plot
  t = t+str_vals.TOFF1;  % "OFF1" period final time
  currRow = currRow+1;    % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_f];  % Add row for Vout ripple plot
  resultData(n,4)=str_vals.TOFF1;   % Save TOFF1 value to table with results
  resultData(n,5)=str_vals.TFB;     % Save TFB value to table with results (it is 0 if there is no feedback in current cycle)
  resultData(n,7)=str_vals.deltaV;  % Save deltaV value to table with results

  % "FEEDBACK" Period
  % Check condition, and if Vout exceed threshold level, calculate run feedback calculations
  if(str_vals.V_EC_f>str_vals.V_EC_Th_Max)

    str_vals = calc_FB(str_vals); % Calculate values for "FEEDBACK" period
    %labelX = t;               % Save coordinate x for label on plot (Optional)
    %labelY = str_vals.V_EC_i; % Save coordinate y for label on plot (Optional)
    currRow = currRow+1;      % next row index for Vout ripple plot
    plotData(currRow,:) = [t,str_vals.V_EC_i];  % Add row for Vout ripple plot
    t = t+str_vals.TFB;       % "FEEDBACK" period final time
    currRow = currRow+1;      % next row index for Vout ripple plot
    plotData(currRow,:) = [t,str_vals.V_EC_f];  % Add row for Vout ripple plot
    resultData(n,5)=str_vals.TFB; % Save TFB value to table with results

  endif

  % "OFF2" Period
  str_vals = calc_OFF2(str_vals); % Calculate values for "OFF2" period
  currRow = currRow+1;  % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_i];  % Add row for Vout ripple plot
  t = t+str_vals.TOFF2; % "OFF2" (AND current cycle) end time
  currRow = currRow+1;    % next row index for Vout ripple plot
  plotData(currRow,:) = [t,str_vals.V_EC_f];  % Add row for Vout ripple plot
  resultData(n,6)=str_vals.TOFF2; % Save TOFF2 value to table with results

  resultData(n,8)=t-tstart; % Save lenght of current cycle to to table with results

endfor

% Display results
format shortg
disp("Cyclo NO. | t, S | TON, S | TOFF1, S | TFB, S | TOFF2, S | deltaV, V | Cycle duration, S");
disp(resultData)

% Plot Vout ripple waveform
plot(plotData(:,1),plotData(:,2));
xlabel("t, S");
ylabel("Output voltage, V");
title("Output ripple");
% text(labelX,labelY, "Shunt regulator is ON"); % Optional


